(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2014 Konstantin Korovin and The University of Manchester. 
    This file is part of iProver - a theorem prover for first-order logic.

        iProver is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
  iProver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
    See the GNU General Public License for more details.
        You should have received a copy of the GNU General Public License
            along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)


open Lib
open Options
open Logic_interface 
(* open Qbf_env*)
open PredElim

(*----- debug modifiable part -----*)

let dbg_flag = false

type dbg_gr = 
  | D_trace
  | D_pred_elim
  | D_final 
  | D_gs
  | D_time

let dbg_gr_to_str = function 
  | D_trace -> "trace"
  | D_pred_elim -> "pred_elim"
  | D_final -> "final"
  | D_gs -> "gs"
  | D_time -> "time"

let dbg_groups =
  [
(*   D_trace;  *)
(*   D_pred_elim; *)
(*   D_final; *)
   D_gs; 
   D_time 
 ]

let module_name = "qbf_preprocess"

(*----- debug fixed part --------*)

let () = Lib.dbg_flag_msg dbg_flag module_name

let dbg group str_lazy =
  Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f =
  Lib.dbg_env_set dbg_flag dbg_groups group f

(*----- debug -----*)


(* can not open both prop modules and Logic_interface since many name clashes *)
module PV  = Prop_var
module PVE = Prop_env
module QE  = Qbf_env 

module PVMap = PV.VMap
module PVSet = PV.VSet

module PLMap = PV.LMap
module PLSet = PV.LSet

type pvar = PV.var
type plit = PV.lit
type qbf_cnf = Qbf_env.qbf_cnf

let bool_type = Symbol.symb_bool_type
let prop_type = Symbol.symb_default_type

(* ------ time_outs ------ *)
exception QBF_prep_timeout

let prep_remaining_time () = 
    ((!current_options.time_out_real*. !current_options.time_out_prep_mult) -. (iprover_running_time ())) 


let prep_qbf_remaining_time () = 
   (((!current_options.time_out_real *. !current_options.time_out_prep_mult)) -. (iprover_running_time ())) 


let prop_subs_time_limit () = (prep_remaining_time ())/.2.

let pred_elim_time_limit () =  (prep_remaining_time ())/.2.

(*  (prep_remaining_time ())/. 2. *)
(*
let get_half_time_limit () = prep_remaining_time ()
*)
  (* prep_qbf_remaining_time ()*)

(*
  let half_time_lim =  (prep_qbf_remaining_time ())/. 2. in
  dbg D_time (lazy 
   (" iprover_running_time: "^(string_of_float (iprover_running_time ()))
  ^" prep_qbf_remaining_time: "^(string_of_float (prep_qbf_remaining_time ()))
  ^" get_half_time_limit: "^(string_of_float half_time_lim)));
  half_time_lim
  *)  


let check_time ~start_time ~time_limit = 
  let current_time = Unix.gettimeofday () in
   if (!current_options.time_out_real > 0.) 
   then 
    (
     let time_diff = (current_time -. start_time) in 
     if (time_diff > time_limit)
     then
       ( 
(*
         dbg D_time (lazy (" timeout time_limit: "^(string_of_float time_limit)
                           ^" iprover_running_time: "^(string_of_float (iprover_running_time ()))
                           ^" prep_qbf_remaining_time: "^(string_of_float (prep_qbf_remaining_time ()))
                           ^" get_half_time_limit: "^(string_of_float (get_half_time_limit ()))));
*)
         raise QBF_prep_timeout
        )
    )     


type prop_fof_env = 
    {
     mutable prop_to_fof_symbs : symbol PVMap.t;
     mutable fof_to_prop_symbs : PV.var SMap.t;

     mutable prop_to_fof_lits : term PLMap.t;
     mutable fof_to_prop_lits : plit TMap.t;

     mutable prop_to_fof_clauses : Clause.clause PVE.CMap.t;
     mutable fof_to_prop_clauses : PVE.clause CMap.t;
   }

let pf_env_create () = 
  {
   prop_to_fof_symbs = PVMap.empty;
   fof_to_prop_symbs = SMap.empty;
   prop_to_fof_lits = PLMap.empty;
   fof_to_prop_lits = TMap.empty;
   prop_to_fof_clauses = PVE.CMap.empty;
   fof_to_prop_clauses = CMap.empty;
 }






let prop_var_to_fsymb pf_env pvar = 
  try 
    PVMap.find pvar pf_env.prop_to_fof_symbs
  with 
    Not_found -> 
      let pred_name = ("$$iProver_qbf_prop_"^(PV.var_to_string pvar)) in 
      let stype = create_stype [] bool_type in
      let pred_symb = create_symbol pred_name stype in
      pf_env.prop_to_fof_symbs <- PVMap.add pvar pred_symb pf_env.prop_to_fof_symbs;
      pf_env.fof_to_prop_symbs <- SMap.add pred_symb pvar pf_env.fof_to_prop_symbs;
      pred_symb

(* can raise Not_found *)
let fof_to_prop_symb pf_env symb = 
  SMap.find symb pf_env.fof_to_prop_symbs

let prop_lit_to_fof pf_env plit = 
  try 
    PLMap.find plit pf_env.prop_to_fof_lits 
  with 
    Not_found -> 
      let (sign,pvar) = PV.lit_to_var plit in
      let pred_symb = prop_var_to_fsymb pf_env pvar in 
      let fof_lit = add_lit_symb sign pred_symb [] in
      let fof_compl_lit = add_compl_lit fof_lit in 
      let compl_plit = PV.compl_lit plit in
      (* add both lit and its complement *) 
      pf_env.prop_to_fof_lits <- PLMap.add plit fof_lit pf_env.prop_to_fof_lits;
      pf_env.prop_to_fof_lits <- PLMap.add compl_plit fof_compl_lit pf_env.prop_to_fof_lits;

      pf_env.fof_to_prop_lits <- TMap.add fof_lit plit pf_env.fof_to_prop_lits;   
      pf_env.fof_to_prop_lits <- TMap.add fof_compl_lit compl_plit pf_env.fof_to_prop_lits;   

(*
  pf_env.prop_to_fof_lits <- PLMap.add plit fof_lit pf_env.prop_to_fof_lits;
  pf_env.fof_to_prop_lits <- TMap.add fof_lit plit pf_env.fof_to_prop_lits;   
 *)      
      fof_lit

        
let fof_lit_to_prop pf_env flit = 
  try 
    TMap.find flit pf_env.fof_to_prop_lits
  with 
    Not_found -> failwith ("fof_lit_to_prop: fof lit "^(Term.to_string flit)^" is not in pf_env")



let prop_clause_to_fof pf_env pclause =   
  let plits = PVE.clause_get_lits pclause in
  let fof_lits = List.map (prop_lit_to_fof pf_env) plits in
  let tstp_source = Clause.tstp_source_input "qbf_preprocess" "qbf" in (* TODO change tstp_source_input *)
  let new_fof_clause = create_clause tstp_source fof_lits in 
  new_fof_clause
    

let fof_clause_to_prop pf_env fclause = 
  let flits = Clause.get_literals fclause in 
  let plits = List.map (fof_lit_to_prop pf_env) flits in 
  let new_prop_clause = PVE.clause_create plits PVE.P_Input in (* TODO change source *)
  new_prop_clause
    

(* slow for some reason
   let prop_clause_to_fof pf_env pclause = 
   try 
   PVE.CMap.find pclause pf_env.prop_to_fof_clauses
   with 
   Not_found ->
   (
   let plits = PVE.clause_get_lits pclause in
   let fof_lits = List.map (prop_lit_to_fof pf_env) plits in
   let tstp_source = Clause.tstp_source_input "qbf_preprocess" "qbf" in (* TODO change tstp_source_input *)
   let new_fof_clause = create_clause tstp_source fof_lits in 
   pf_env.prop_to_fof_clauses <- PVE.CMap.add pclause new_fof_clause pf_env.prop_to_fof_clauses ;
   new_fof_clause
   )

   let fof_clause_to_prop pf_env fclause = 
   try
   CMap.find fclause pf_env.fof_to_prop_clauses    
   with 
   Not_found ->
   (
   let flits = Clause.get_literals fclause in 
   let plits = List.map (fof_lit_to_prop pf_env) flits in 
   let new_prop_clause = PVE.clause_create plits PVE.P_Input in (* TODO change source *)
   pf_env.fof_to_prop_clauses <- CMap.add fclause new_prop_clause pf_env.fof_to_prop_clauses;
   new_prop_clause
   )
 *)
(*-----------preliminaries -----------------------*)

let get_max_var_lvl qbf c =
  let lits = PVE.clause_get_lits c in 
  assert (not (lits == []));
  let get_lvl_lit lit = 
    let v = PV.get_var_lit lit in 
    QE.qbf_get_var_level qbf v
  in
  let cmp_lit_lvl l1 l2 = Pervasives.compare (get_lvl_lit l1) (get_lvl_lit l2) in
  let max_lvl_lit = list_find_max_element cmp_lit_lvl lits in
  get_lvl_lit max_lvl_lit 

(* based on lvl weaker than based on dep *)
let not_to_elim_var_set qbf =     
  let f f_nte_rest_set c = 
    let lits = PVE.clause_get_lits c in 
    let max_lvl = get_max_var_lvl qbf c in 
    let g g_nte_rest_set lit =
      let v = PV.get_var_lit lit in
      let (lvl, quant) = QE.qbf_get_var_qlevel qbf v in          
      if (quant = QE.E) && (lvl < max_lvl) 
      then 
        PVSet.add v g_nte_rest_set
      else
        g_nte_rest_set
    in 
    List.fold_left g f_nte_rest_set lits
  in
  List.fold_left f PVSet.empty qbf.QE.qbf_clauses

(*can raise Not_found *)
let get_max_a_var_lvl_dep qbf var_dep c =
  let lits = PVE.clause_get_lits c in 
  assert (not (lits == []));
  let a_dep_var_set = 
    let f a_set lit = 
      let var = PV.get_var_lit lit in
      let (_lvl, quant) = QE.qbf_get_var_qlevel qbf var in
      match quant with 
      |QE.A -> PVSet.add var a_set 
      |QE.E -> 
          let v_a_set = QE.get_var_dep var_dep var in 
          PVSet.union v_a_set a_set
    in
    List.fold_left f PVSet.empty lits
  in
  let a_dep_list = PVSet.elements a_dep_var_set in 
  let get_lvl_var v = QE.qbf_get_var_level qbf v
  in
  let cmp_lvl_var v1 v2 = Pervasives.compare (get_lvl_var v1) (get_lvl_var v2) in
  let max_lvl_var = list_find_max_element cmp_lvl_var a_dep_list in
  get_lvl_var max_lvl_var 


let not_to_elim_var_set_dep qbf var_dep =     
  let f f_nte_rest_set c = 
    let lits = PVE.clause_get_lits c in 
    try 
      let max_lvl = get_max_a_var_lvl_dep qbf var_dep c in 
      let g g_nte_rest_set lit =
        let v = PV.get_var_lit lit in
        let (lvl, quant) = QE.qbf_get_var_qlevel qbf v in          
        if (quant = QE.E) && (lvl < max_lvl) 
        then 
          PVSet.add v g_nte_rest_set
        else
          g_nte_rest_set
      in
      List.fold_left g f_nte_rest_set lits
    with 
      Not_found -> f_nte_rest_set (* No a/a_dep_vars *)
  in
  List.fold_left f PVSet.empty qbf.QE.qbf_clauses


let to_elim_symb_set qbf pf_env = 
  let nte_var_set = not_to_elim_var_set qbf in
  let f pvar symb elim_set = 
    let (_lvl, quant) = QE.qbf_get_var_qlevel qbf pvar in 
    if (quant = QE.E) && (not (PVSet.mem pvar nte_var_set))
    then
      SSet.add symb elim_set
    else 
      elim_set 
  in
  PVMap.fold f pf_env.prop_to_fof_symbs SSet.empty

let all_e_symb_set qbf pf_env = 
  let f pvar symb e_set = 
    if (QE.qbf_is_e_var qbf pvar) 
    then 
      SSet.add symb e_set
    else 
      e_set 
  in
  PVMap.fold f pf_env.prop_to_fof_symbs SSet.empty

let prop_subsumption qbf pf_env = 
  let fof_clauses = List.map (prop_clause_to_fof pf_env) qbf.QE.qbf_clauses in 
  let f c = 
    let simp_c =  Simplify.forward_prop_subsume c in 

    if (not (c == simp_c))
    then 
      (
       dbg D_gs (lazy (" before: "^(Clause.to_string c)));
       dbg D_gs (lazy (" after: "^(Clause.to_string simp_c)));
       let pclause = fof_clause_to_prop pf_env simp_c in 
       let new_pclause = QE.q_resolve_clause qbf pclause in 
       if (not (new_pclause == pclause)) 
       then 
         prop_clause_to_fof pf_env new_pclause 
       else 
         simp_c
      )
    else
      c
  in
  let simp_fof_clauses = List.map f fof_clauses in 
  let new_prop_clauses = List.map (fof_clause_to_prop pf_env) simp_fof_clauses in 
  qbf.QE.qbf_clauses <- new_prop_clauses 


let get_num_of_plits cl_list =
  List.fold_left (fun rest c -> (List.length (PVE.clause_get_lits c)) + rest) 0 cl_list 
    
let prop_subsumption_exhaustive_clause clause = 
  let finished = ref false in
  let simp_clause_ref = ref clause in
  let count = ref 0 in 
  while (not !finished)
  do
    let simp_c = Simplify.forward_prop_subsume !simp_clause_ref in 
    (
    if (not (!simp_clause_ref == simp_c))
    then
      (
       count:=!count+1;
       dbg D_gs (lazy (" before: "^(Clause.to_string !simp_clause_ref)));
       dbg D_gs (lazy (" after: "^(Clause.to_string simp_c)^" count: "^(string_of_int !count))); 
       simp_clause_ref := simp_c;              
      )
    else 
      (
       finished:=true;
      ) 
    )
  done;
  !simp_clause_ref

let cmp_pclause_num_lits c1 c2 = 
  let num_lits1 = List.length  (PVE.clause_get_lits c1) in
  let num_lits2 = List.length  (PVE.clause_get_lits c2) in
  Pervasives.compare num_lits1 num_lits2

let prop_subsumption_exhaustive qbf pf_env = 
  let start_time = Unix.gettimeofday () in 
  let time_limit = prop_subs_time_limit () in
  dbg D_time (lazy (" prop_subsumption_exhaustive: time_limit: "^(string_of_float time_limit)));
  print_string (" ps_s  "); flush stdout;
  let num_of_gs_subs_before = Statistics.get_val_stat Statistics.prop_fo_subsumed in
  let sorted_pclauses =  List.sort cmp_pclause_num_lits qbf.QE.qbf_clauses in
  let fof_clauses = List.map (prop_clause_to_fof pf_env) sorted_pclauses in 
  let changed = ref true in
  let simp_clauses = ref fof_clauses in
 
 (* (try*)
    while !changed 
    do
      let new_changed =  ref false in 
      let g c =
        try
          check_time start_time time_limit;
          let simp_c = prop_subsumption_exhaustive_clause c in 
          if (not (c == simp_c))
          then
            (
             dbg D_gs (lazy (" before: "^(Clause.to_string c)));
             dbg D_gs (lazy (" after: "^(Clause.to_string simp_c))); 
             new_changed:=true;
             let pclause = fof_clause_to_prop pf_env simp_c in 
             let new_pclause = QE.q_resolve_clause qbf pclause in 
             if (not (new_pclause == pclause)) 
             then 
               prop_clause_to_fof pf_env new_pclause 
             else 
               simp_c
            )
          else
            c   
        with 
          QBF_prep_timeout -> c
      in
      let new_simp_cls = List.map g !simp_clauses in
      changed := !new_changed;
      simp_clauses := new_simp_cls;
    done;
  let new_prop_clauses = List.map (fof_clause_to_prop pf_env) !simp_clauses in 
(* statistics out *)
  let num_of_gs_subs_after = Statistics.get_val_stat Statistics.prop_fo_subsumed in
  let end_time =  Unix.gettimeofday () in 
  print_string 
    (
     (string_of_int (num_of_gs_subs_after - num_of_gs_subs_before))
     ^"c "
     ^(string_of_int (truncate (end_time -. start_time))^"s ps_e" )); flush stdout;
(* end statistics out *)
  qbf.QE.qbf_clauses <- new_prop_clauses

(*
      (let new_prop_clauses = List.map (fof_clause_to_prop pf_env) !simp_clauses in 
      qbf.QE.qbf_clauses <- new_prop_clauses;
      raise QBF_prep_timeout 
      )
*)

(*
  let prop_subsumption_exhaustive_clause clause = 
  let rec f c = 
  let simp_c = Simplify.forward_prop_subsume c in 
  if (not (c == simp_c))
  then
  (dbg D_gs (lazy (" before: "^(Clause.to_string c)));
  dbg D_gs (lazy (" after: "^(Clause.to_string simp_c))); 
  f simp_c 
  )
  else 
  c 
  in
  f clause

  let prop_subsumption_exhaustive qbf pf_env = 
  let fof_clauses = List.map (prop_clause_to_fof pf_env) qbf.QE.qbf_clauses in 
  let rec f changed cls = 
  if (not changed) 
  then
  cls
  else
  (
  let new_changed =  ref false in 
  let g c =
  let simp_c = prop_subsumption_exhaustive_clause c in 
  if (not (c == simp_c))
  then
  (
  new_changed:=true;
  let pclause = fof_clause_to_prop pf_env simp_c in 
  let new_pclause = QE.q_resolve_clause qbf pclause in 
  if (not (new_pclause == pclause)) 
  then 
  prop_clause_to_fof pf_env new_pclause 
  else 
  simp_c
  )
  else
  c               
  in
  let simp_cls = List.map g cls in 
  f !new_changed simp_cls
  )
  in
  let simp_fof_clauses = f true fof_clauses in
  let new_prop_clauses = List.map (fof_clause_to_prop pf_env) simp_fof_clauses in 
  qbf.QE.qbf_clauses <- new_prop_clauses 
 *)

(*
  let prop_subsumption_exhaustive qbf pf_env = 
  let exhausted = ref false in
  while (not !exhausted)
  do
  let old_num_of_lits = get_num_of_plits qbf.QE.qbf_clauses in
  prop_subsumption qbf pf_env;
  let new_num_of_lits = get_num_of_plits qbf.QE.qbf_clauses in
  (if (old_num_of_lits = new_num_of_lits) 
  then 
  exhausted:= true
  );
  done
 *)

(*-----------------------------------*)
let pred_elim_qbf qbf pf_env var_dep = 
  
  if (not !current_options.qbf_pred_elim) 
  then 
    ()
  else   
    begin
      let fof_clauses = List.map (prop_clause_to_fof pf_env) qbf.QE.qbf_clauses in 


(*      
        let num_of_input_clauses = List.length fof_clauses in 

        let init_pred_elim_set = (* inner existential quantifier block *)        
        let (_lvl,qb) = QE.qbf_get_max_lvl_qb qbf in
        assert (qb.QE.qb_quant == QE.E);
        let pvar_set = qb.QE.qb_vars in
        let fsymb_set = 
        let f pvar curr_fset = 
        let symb = prop_var_to_fsymb pf_env pvar in 
        SSet.add symb curr_fset
        in
        PVSet.fold f pvar_set SSet.empty
        in
        fsymb_set
        in
 *)
(*
  let init_pred_elim_set = (* all exist vars*)
  let f pvar fsymb elim_symb_set = 
  let (lvl, quant) = QE.qbf_get_var_qlevel qbf pvar in          
  if (quant = QE.E) 
  then 
  let dep_set = 
  try
  PVMap.find pvar var_dep
  with 
  Not_found -> PVSet.empty (* failwith "init_pred_elim_set var is not in var_dep"*)
  in
  let greater_a_lvl_test a_var = 
  let a_lvl = QE.qbf_get_var_level qbf a_var in 
  a_lvl > lvl 
  in
  if (PVSet.exists greater_a_lvl_test dep_set) 
  then (* do not eliminate *)
  elim_symb_set
  else
  SSet.add fsymb elim_symb_set 
  else
  elim_symb_set
  in
  let elim_set = PVMap.fold f pf_env.prop_to_fof_symbs SSet.empty in
  dbg_env D_pred_elim 
  (fun () -> 
  dbg D_pred_elim (lazy " variables for elimination:\n");
  out_str (list_to_string Symbol.to_string  (SSet.elements elim_set) " ");
  dbg D_pred_elim (lazy "\n");
  );
  elim_set
  in
 *)

      let init_pred_elim_set = all_e_symb_set qbf pf_env in
      
(* works but does not simplify clauses with non-eligible e symbols

   let init_pred_elim_set = 
   let elim_set = to_elim_symb_set qbf pf_env in
   dbg_env D_pred_elim 
   (fun () -> 
   dbg D_pred_elim (lazy " variables for elimination:\n");
   out_str (list_to_string Symbol.to_string  (SSet.elements elim_set) " ");
   dbg D_pred_elim (lazy "\n");
   );
   elim_set
   in 
 *)    

      let not_to_elim_pvar_set =  not_to_elim_var_set_dep qbf var_dep  (* not_to_elim_var_set qbf *) in

      
(*
  let num_of_input_clauses = (Statistics.get_val_stat num_of_input_clauses) in
 *)
      let pred_elim_options = 
        {
	 pe_has_eq = false; 
	 pe_estim_num_of_lits =  2000; (* 10000;*)

(*
  (if (is_ver_epr ())
  then
  if  (num_of_input_clauses < 1000000 (* 100 000 *))
  then
  45
  else
  25  (*260*) (* 25 *) (* 45 *) (* 65 *) (*85*) (* 65 *)
  else
  40 (* 1000 *)
  );
 *)

	 pe_conclusion_limit_test = 
         (fun c -> 
           (Clause.num_of_symb c) < 1000
         );

         pe_preprocess_conclusion_extern = 
         (fun c -> 

           let pclause = fof_clause_to_prop pf_env c in 
           let new_pclause = QE.q_resolve_clause qbf pclause in 
           if (not (new_pclause == pclause)) 
           then 
             prop_clause_to_fof pf_env new_pclause 
           else 
             c

         );
(*
  pe_clause_length_limit = 
  (if  (is_ver_epr ())
  then
  20  (* 8 *) (* 26 *) (* 13*)
  else
  50
  ); 
 *)
(*100;*)
(*	(if  (is_ver_epr ())
  then
  13  (* 8 *) (* 26 *) (* 13*)
  else
  8 (* 8 *) (* 50 *)
  );
 *)
(* not evaluated on aigs *)
	 pe_keep_elim =  
         (fun  ~elim_symb ~clauses_before_elim ~clauses_after_elim ->           
           let pelim_var = fof_to_prop_symb pf_env elim_symb in 
           if (PVSet.mem pelim_var not_to_elim_pvar_set) 
           then 
             false 
           else

             let num_cl_before = (List.length clauses_before_elim) in 
             let num_cl_after  = (List.length clauses_after_elim) in
             let keep_res =            
               if num_cl_after = 0
               then 
                 ( 
                   dbg D_pred_elim (lazy ("pe_keep_elim: all clauses are simplified" ));
                   true 
                  )
               else 
                 begin
                   assert (num_cl_before > 1);
                   
(*
  let get_num_of_symb cl_list = 
  List.fold_left (fun rest c -> (Clause.num_of_symb c) + rest) 0 cl_list 
  in
 *) 
                   let get_num_of_lits cl_list =
                     List.fold_left (fun rest c -> (Clause.length c) + rest) 0 cl_list 
                   in 
                   let get_num_e_lits_clause cl = 
                     let f rest lit = 
                       let plit = fof_lit_to_prop pf_env lit in
                       let pvar = PV.get_var_lit plit in
                       if (QE.qbf_is_e_var qbf pvar) 
                       then 
                         rest+1 
                       else
                         rest
                     in
                     Clause.fold f 0 cl 
                   in
                   let get_num_e_lits cl_list = 
                     List.fold_left (fun rest c -> (get_num_e_lits_clause c) + rest) 0 cl_list 
                   in
(* 
   List.fold_left (fun rest c -> (Clause.length c) + rest) 0 cl_list 
 *)

                   let num_lits_before = get_num_of_lits clauses_before_elim in 
                   let num_lits_after = get_num_of_lits clauses_after_elim in 
                   
                   let num_e_lits_before = get_num_e_lits clauses_before_elim in 
                   let num_e_lits_after  = get_num_e_lits clauses_after_elim in 
                   
                   let cl_max_lits_before = list_find_max_element Clause.cmp_num_lits clauses_before_elim in 
                   let cl_max_lits_after = list_find_max_element Clause.cmp_num_lits clauses_after_elim in 
                   
                   let cl_max_e_lits_before_num = get_num_e_lits_clause cl_max_lits_before in 
                   let cl_max_e_lits_after_num =  get_num_e_lits_clause cl_max_lits_after in
                   
(*TODO: add varible square mesure per clause cl = \sum (num var lits)^2 *)
                   
                   dbg D_pred_elim (lazy ("pe_keep_elim: num_cl: before: "^(string_of_int num_cl_before)
                                          ^" after: "^(string_of_int num_cl_after)));
                   
                   dbg D_pred_elim (lazy ("pe_keep_elim: num_lits: before:  "^(string_of_int num_lits_before)
                                          ^" after: "^(string_of_int num_lits_after)));

                   dbg D_pred_elim (lazy ("pe_keep_elim: num_e_lits: before:  "^(string_of_int num_e_lits_before)
                                          ^" after: "^(string_of_int num_e_lits_after)));

                   dbg D_pred_elim (lazy ("pe_keep_elim: cl_max_e_lits: before:  "^(string_of_int cl_max_e_lits_before_num)
                                          ^" after: "^(string_of_int cl_max_e_lits_after_num)));
                   
                   (*----------------*)     

(*                  num_lits_after <=  num_lits_before
                    &&
 *)

                   num_e_lits_after <= num_e_lits_before                     
                     &&
                   ( num_e_lits_after <= !current_options.qbf_split
                   ||
                     cl_max_e_lits_after_num <= cl_max_e_lits_before_num
                    )
                 end
             in
             dbg D_pred_elim (lazy ("pe_keep_elim: keep ?: "^(string_of_bool keep_res)));            
             keep_res               
         ); 
         
	 (* (if (is_ver_epr ())   *)
	 (* then *)
	 (*   (fun ~num_cl_before ~num_cl_after ->  *)
	 (*     (num_cl_after <= num_cl_before)  || (num_cl_after <= 2)) (\* (num_cl_after <= 6) ) *\) *)
	 (* else *)
	 (*   (fun ~num_cl_before ~num_cl_after ->  *)
	 (*     (num_cl_after <= num_cl_before) (\* ||  (num_cl_after <= 6)*\) ) *)
	 (* ); *)

	 pe_elim_order_cmp_fun = (fun _context s1 s2 ->  Symbol.compare s1 s2); (*pred_elim_cmp_fun;*) (* NOT USED *)
	 pe_elimination_set = init_pred_elim_set; 

(* 1/4 of the remaining time *)
         pe_time_limit =
         
         (if (!current_options.time_out_real > 0.)
         then  
           let time_limit = pred_elim_time_limit () in 
           if (time_limit > 0.) 
           then              
             (
              dbg D_time (lazy (" pred_elim timelimit "^(string_of_float time_limit)));
              time_limit 
             )
           else
             1.
         else
           -1.
         );	
(* simplifications *)

(*	subs_cl_to_cl_limit = 100000; *)
	 subs_cl_to_cl_limit = 100000;  


(* sim prop *)
	 prop_glb_subs =  (* false; *)   true; 
(*
  (if (num_of_input_clauses < 100000 )
  then
  true  (* prop global subsumtion changed for finite models exp. *)
  else
  false
  );
 *)	
(* sim local *) 
	 
         lcl_add_to_sub_index_test = (fun _c -> false);
(*
  (fun c ->
  (((Clause.num_of_var c) <= 100) && ((Clause.num_of_symb c) <= 1000))
  );
 *)
	 (* Options.res_subs_type: type res_subs_type = Subs_Full | Subs_Subset | Subs_By_Length of int *)
	 
	 lcl_fwd_subs =   false; (* true; *)
(*
  (if (num_of_input_clauses < 200000)
  then
  true
  else
  false
  );
 *)      
	 lcl_fwd_subs_res =  false; (* true;*)
(*
  (
  if (num_of_input_clauses < 100000 (* 100000*))
  then
  true
  else
  false
  else
  true (*  false *)
  );
 *)
	 
	 lcl_bwd_subs = Subs_Subset;
         
(* Subs_Full; *)
(*  Subs_By_Length(2); *)
(*
  (if (is_ver_epr ())
  then
  if  (num_of_input_clauses < 100000) (* (num_of_input_clauses < 400000)*)
  then
  Subs_By_Length(1) (* Subs_By_Length(1) *)
  else
  Subs_Subset
  
  (* Subs_Subset*)  (*(Subs_By_Length(1)) *) (* Subs_Full *)
  else
  (* Subs_Full*)  Subs_By_Length(20) (* Subs_Subset *)(* (Subs_By_Length(2)) *) (* (Subs_Full) *)
  );
 *)
	 lcl_bwd_subs_res = Subs_Subset;
(*
  lcl_bwd_subs_res =   (* Subs_Full;*)
  (if  (is_ver_epr ())
  then
  if (num_of_input_clauses < 100000 (* 100000*))
  then
  Subs_By_Length(1)
  else
  Subs_Subset
  
  (* Subs_Subset*) (*  (Subs_By_Length(1)) *)(* Subs_Full *)
  else
  (* Subs_Full *)  (Subs_By_Length(20))  (* Subs_Subset *)(* (Subs_By_Length(2))*) (* (Subs_Full) *)
  );
 *)
	 
(* sim global *)

         glb_add_to_sub_index_test =   (fun _c -> false);  (* (fun c -> (Clause.length c) <= 200); *)
(*
  (fun c ->
  ((Clause.num_of_var c) <= 100)  && ((Clause.num_of_symb c) <= 1000));
 *)	
	 glb_fwd_subs =   false; (* true; *)
(*
  (if (is_ver_epr ())
  then
  if (num_of_input_clauses < 100000 (*200000 *))
  then
  true
  else
  false
  else
  true  (* false *)
  );
 *)	
	 glb_fwd_subs_res = false; (*false;*) (* true; *)
(*	(if (is_ver_epr ())
  then
  if (num_of_input_clauses < 100000 (* 100000 *) )
  then
  true
  else
  false
  else
  true  (* false; *) 
  );
 *)
	 glb_bwd_subs = (* Subs_By_Length(10);*) Subs_Subset; (*Subs_By_Length(100);*)

(* (* Subs_Full; *)
   (if (is_ver_epr ())
   then
   if (num_of_input_clauses < 100000 (*200000*))
   then
   Subs_By_Length(1)
   else
   Subs_Subset (*  Subs_By_Length(1) *)  (*Subs_Full*)
   else
   (* Subs_Full; *) (* Subs_Subset *)  Subs_By_Length(10);  (* Subs_By_Length(4); *) 
   );
 *)	
	 glb_bwd_subs_res = (* Subs_By_Length(10);*)  Subs_Subset; (*Subs_By_Length(100);*) (* Subs_Full;*)
(*
  (if (is_ver_epr ())
  then
  if (num_of_input_clauses < 100000 (* 200000 *))
  then
  Subs_By_Length(1)
  else
  Subs_Subset (* Subs_By_Length(1) *)(* Subs_Full*)
  else
  (* Subs_Full; *) (* Subs_Subset *) Subs_By_Length(10); (* Subs_By_Length(4); *) (*Subs_Full; *)
  );
 *)
       }
      in
      (*let fixed_point_reached = ref false in*)
      (* out_str (Clause.clause_list_to_string !current_list); *)

      let new_fof_clauses = PredElim.predicate_elimination pred_elim_options (Clause.CL_List (fof_clauses)) in
      let new_prop_clauses = List.map (fof_clause_to_prop pf_env) new_fof_clauses in 
      qbf.QE.qbf_clauses <- new_prop_clauses;    
    end


let get_var_dep qbf = 
  let var_dep_map = 
    if !current_options.qbf_sk_in 
    then 
      QE.qbf_inner_var_dep qbf
    else
      QE.qbf_outer_var_dep qbf
  in
  var_dep_map


(* remove all vars from prefix that are not in cnf *)
let reduce_vars_qbf qbf =   
  let get_var_set cl_list = 
    let f f_v_set c = 
      let g g_v_set lit = 
        PVSet.add (PV.get_var_lit lit) g_v_set 
      in
      List.fold_left g f_v_set (PVE.clause_get_lits c)
    in
    List.fold_left f PVSet.empty cl_list
  in
  let new_qbf_vars = get_var_set qbf.QE.qbf_clauses in 
  let reduce_qp lvl qb new_qp = 
    let new_qb_vars = PVSet.inter qb.QE.qb_vars new_qbf_vars in 
    if (PVSet.is_empty new_qb_vars)
    then
      new_qp
    else
      let new_qb =  {qb with QE.qb_vars = new_qb_vars } in 
      IntMap.add lvl new_qb new_qp
  in
  let new_qbf_pref = IntMap.fold reduce_qp qbf.QE.qbf_pref IntMap.empty in 
  let new_qbf_var_level = 
    let f v lvlq new_vl = 
      if (PVSet.mem v new_qbf_vars) 
      then
        PVMap.add v lvlq new_vl
      else
        new_vl
    in
    PVMap.fold f qbf.QE.qbf_var_level PVMap.empty 
  in 
  qbf.QE.qbf_pref <- new_qbf_pref;
  qbf.QE.qbf_var_level <- new_qbf_var_level;
  qbf.QE.qbf_vars <- new_qbf_vars


let preprocess_qbf qbf = 
  print_string ((s_pref_str ())^"QBF preprocessing...");
  flush stdout;
  dbg_env D_final 
    (fun() -> 
      dbg D_final (lazy "");
      dbg D_final (lazy "--------Init var_dep-------\n\n");
      QE.out_var_dep (get_var_dep qbf);
      dbg D_final (lazy "--------Init QBF-------\n\n");
      QE.out_qbf qbf;
    );
  let (taut, non_taut) = List.partition PVE.is_tautology qbf.QE.qbf_clauses in 
  Statistics.incr_int_stat (List.length taut) Statistics.qbf_num_tautologies;
  qbf.QE.qbf_clauses <- non_taut;
  QE.qbf_q_resolve qbf;  

  let pf_env = pf_env_create () in 

(*  try  
  begin
*)
  prop_subsumption_exhaustive qbf pf_env;
  
  let repeat_pred_elim = ref true in 
  let cycle_num = ref 0 in 

  QE.qbf_split_clauses 64 qbf; 
      
  while (!repeat_pred_elim &&  (prep_qbf_remaining_time () > -5.))
  do
    cycle_num:=!cycle_num + 1;
    Statistics.incr_int_stat 1 Statistics.qbf_prep_cycles;
    dbg D_pred_elim (lazy (" start new cycle: "^(string_of_int !cycle_num)));
    dbg D_pred_elim (lazy (" start new cycle stat: "^(string_of_int (Statistics.get_val_stat Statistics.qbf_prep_cycles))));

    let old_num_of_clauses = List.length qbf.QE.qbf_clauses in
    let old_num_of_lits = get_num_of_plits qbf.QE.qbf_clauses in


(*    let old_num_of_elim = Statistics.get_val_stat Statistics.pred_elim in*)
    let var_dep = get_var_dep qbf in
    dbg_env D_pred_elim 
      (fun () -> 
        QE.out_var_dep var_dep;
      );

    pred_elim_qbf qbf pf_env var_dep;    
        
(*    prop_subsumption qbf pf_env; *)
    prop_subsumption_exhaustive qbf pf_env; 

(*    reduce_vars_qbf qbf;    *) (* TODO: fix Not_found on BLOCKS4ii.7.2.qdimacs after reduction *)
    let new_num_of_clauses = List.length qbf.QE.qbf_clauses in
    let new_num_of_lits = get_num_of_plits qbf.QE.qbf_clauses in
(*    let new_num_of_elim = Statistics.get_val_stat Statistics.pred_elim in*) (* TODO: fix statistics *)

        (
         if ((old_num_of_lits = new_num_of_lits) && (old_num_of_clauses = new_num_of_clauses))
             (* ((old_num_of_clauses = new_num_of_clauses))*)
         then 
           repeat_pred_elim := false 
         else
           ()
        )
      done;

  out_str "";

(*    end 
  with 
    QBF_prep_timeout ->  *)
(*      begin *)
  QE.qbf_split_clauses !current_options.qbf_split qbf;
  dbg_env D_final 
    (fun() -> 
      dbg D_final (lazy "--------Final var_dep-------\n\n");
      QE.out_var_dep (get_var_dep qbf);
      dbg D_final (lazy "--------Final QBF-------\n\n");
      QE.out_qbf qbf;
    )
(*      end *)
        
